#include <stdlib.h>
#include <windows.h>

void redraw(void);
LRESULT CALLBACK proc_msg(HWND win, unsigned int msg, WPARAM wparam, LPARAM lparam);

HWND win;
HDC win_dc, fb_dc;
HBITMAP hbm_fb;
int fb_pitch, fb_bpl;

int WINAPI WinMain(HINSTANCE hinst, HINSTANCE hprev, LPSTR cmdline, int showcmd)
{
	int i, j;
	unsigned char *pixels, *pptr;
	MSG msg;
	BITMAP bm;
	BITMAPINFO *bmi;
	unsigned char *bmibuf;
	WNDCLASS wc = {0};

	wc.lpszClassName = "foowin";
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = proc_msg;
	wc.hInstance = hinst;
	wc.hIcon = LoadIcon(0, IDI_APPLICATION);
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);

	if(!RegisterClass(&wc)) {
		MessageBox(0, "Failed to register window class", "error", MB_OK);
		return 1;
	}

	if(!(win = CreateWindow("foowin", "foo", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT,
			CW_USEDEFAULT, 320, 200, 0, 0, hinst, 0))) {
		MessageBox(0, "Failed to create window", "error", MB_OK);
		UnregisterClass("foowin", hinst);
		return 1;
	}
	ShowWindow(win, 1);

	win_dc = GetDC(win);
	fb_dc = CreateCompatibleDC(win_dc);
	hbm_fb = CreateCompatibleBitmap(win_dc, 320, 200);
	SelectObject(fb_dc, hbm_fb);
	GetObject(hbm_fb, sizeof bm, &bm);

	fb_pitch = bm.bmWidthBytes;
	fb_bpl = bm.bmPlanes;

	if(!(pixels = malloc(320 * 200 / 2))) {
		MessageBox(0, "Failed to allocate pixel buffer", "error", MB_OK);
		DestroyWindow(win);
		UnregisterClass("foowin", hinst);
		return 1;
	}

	pptr = pixels;
	for(i=0; i<200; i++) {
		for(j=0; j<320; j++) {
			int xor = (i ^ j) & 0xf;
			*pptr++ = (xor & 0xf) | (xor << 4);
		}
	}

	bmibuf = calloc(1, sizeof *bmi + 16 * sizeof(RGBQUAD));
	bmi = (BITMAPINFO*)bmibuf;
	bmi->bmiHeader.biSize = sizeof bmi->bmiHeader;
	bmi->bmiHeader.biWidth = 320;
	bmi->bmiHeader.biHeight = 200;
	bmi->bmiHeader.biPlanes = 1;
	bmi->bmiHeader.biBitCount = 4;
	bmi->bmiHeader.biCompression = BI_RGB;

	for(i=0; i<16; i++) {
		bmi->bmiColors[i].rgbRed = i;
		bmi->bmiColors[i].rgbGreen = i;
		bmi->bmiColors[i].rgbBlue = i;
		bmi->bmiColors[i].rgbReserved = 0;
	}

	/*SetBitmapBits(hbm_fb, 320 * 200 / 2, pixels);*/
	SetDIBits(fb_dc, hbm_fb, 0, 200, pixels, bmi, DIB_PAL_COLORS);

	while(GetMessage(&msg, 0, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	/*
	for(;;) {
		while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
			if(msg.message == WM_QUIT) goto done;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		redraw();
	}
	*/

done:
	free(pixels);
	UnregisterClass("foowin", hinst);
	return 0;
}

void redraw(void)
{
	BitBlt(win_dc, 0, 0, 320, 200, fb_dc, 0, 0, SRCCOPY);
}

LRESULT CALLBACK proc_msg(HWND win, unsigned int msg, WPARAM wparam, LPARAM lparam)
{
	switch(msg) {
	case WM_CLOSE:
		DestroyWindow(win);
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	case WM_PAINT:
		redraw();
		ValidateRect(win, 0);
		break;

	case WM_KEYDOWN:
		if(wparam == VK_ESCAPE) {
			DestroyWindow(win);
		}
		break;

	default:
		return DefWindowProc(win, msg, wparam, lparam);
	}
	return 0;
}
